/**
 * Java-side application for profiling a testbed.
 * 
 *
 * @author Chad S Metcalf <cmetcalf@mines.edu>
 * @date December 1 2006
 */

import java.io.*;
import java.util.*;

import net.tinyos.message.*;
import net.tinyos.packet.*;
import net.tinyos.util.*;



public class TestbedProfiler implements MessageListener {

	
    private TreeMap<Integer, Mote> motes = new TreeMap<Integer, Mote>();
    private BufferedWriter out;
    private final boolean tossim = false;
    private final int channel, payload, power, iterations, delay;
    private int nodeCount = 0;
    private String aliases = "";
   
    
    
    public TestbedProfiler(String map, String log, int iter, int pwr, int chan, int size, int ms) {
    	
    	iterations = iter;
    	power      = pwr;
    	channel    = chan;
    	payload    = size;
    	delay      = ms;
    	
    	
    	importMotes(map);
    	createLogfile(log);
    	
    	if (!tossim) {
    		
    		channelSync();
    		
    	}
    	
    	
    	profile();
    	closeLog();
    	
    }
    
    
    private void importMotes(String map) {

    	String temp;
    	
    	
    	try {

    		BufferedReader in = new BufferedReader(new FileReader(map));


    		while ((temp = in.readLine()) != null) {

    			if  ((temp.length() > 0) && !(temp.startsWith("#"))) {

                    String[] cols = temp.split("\\s+");

                    addMote(++nodeCount, cols[0], cols[1]);
                    
                }
    		}


    		in.close();

    	} catch (IOException e) {

    		System.err.println("File input error");

    		e.printStackTrace();

    	}
    }
    
    
    private void createLogfile(String filename) {
    	
    	System.err.println(filename);
    	
    	try {
   
            out = new BufferedWriter(new FileWriter(filename));
            
            out.write(
            	"#nodes "      + nodeCount  + "\n" +
            	"#iterations " + iterations + "\n" +
            	"#power "      + power      + "\n" +
            	"#channel "    + channel    + "\n" +
            	"#payload "    + payload    + "\n" +
            	"#delay "      + delay      + "\n\n"
            );
            
            out.write(aliases);
            
            out.write("\n#recv src seqno rssi lqi\n");
            
            out.flush();

    	} catch (IOException e) {
    		
            e.printStackTrace();
            
            System.exit(1);
            
        }
    }
    
        
    private void addMote(int id, String name, String location) {
    	
    	PhoenixSource phoenix;


    	if (location == null) {
        
    		phoenix = BuildSource.makePhoenix(PrintStreamMessenger.err);
        
    	} else {
        
    		phoenix = BuildSource.makePhoenix(location, PrintStreamMessenger.err);
        
    	}

    	
        MoteIF mif = new MoteIF(phoenix);
        
        
        mif.registerListener(new ReportMessage(), this);
        mif.registerListener(new PrintfMsg(), this);
        
        
    	motes.put(new Integer(id), new Mote(name, mif));
    	
    	
    	try {
			
			if (id == Integer.parseInt(name)) {
		
				return;
			
			}
			
		} catch (NumberFormatException e) { }
		
		
		aliases += "#alias " + id + " " + name +  "\n";
    	
    }
    

    private void channelSync() {
    	
    	System.out.println("Syncing motes on channel " + channel + "...");
    	
    	
    	Iterator<Integer> it = motes.keySet().iterator();

    	
    	try {

    		while (it.hasNext()) {

    			sendCommandPacket(it.next().intValue(), true);

    		}

    		Thread.sleep(2500);

    	} catch (InterruptedException e) {

    		System.err.println("Channel-sync error...");

    	}

    }
    
    
    private void profile() {

    	int sleepTime = delay * iterations + 500;
    	
    	
        if (tossim) {
        	
            for(int i = 0; i < 52; i++) {
            
                sendCommandPacket(i, false);

                
                try {
                
                	Thread.sleep(sleepTime);
                
                } catch (InterruptedException e) {
                
                	System.err.println("Profile interrupted...");
                
                }
            }
            
        } else {
        	
            Iterator<Integer> it = motes.keySet().iterator();
            
            
            while (it.hasNext()) {
            	
                sendCommandPacket(it.next().intValue(), false);

                
                try {
                	
                    Thread.sleep(sleepTime);

                } catch (InterruptedException e) {
                
                	System.err.println("Profile interrupted...");
                
                }
            }
        }

        
        try {
        	
            out.flush();
        
        } catch (IOException e) {
        
        	System.err.println("Problems flushing log...");
        
        }
    }
    

    private void sendCommandPacket(int id, boolean sync) {

    	try {

            Mote mote = null;
            Integer moteID = new Integer(id);

            
            CommandMessage cmd = new CommandMessage();
            	cmd.set_moteID(id);
            	cmd.set_iterations(iterations);
            	cmd.set_interval(delay);
            	cmd.set_power((short) power);
            	cmd.set_bytes((short) payload);
            	cmd.set_channel((short) (sync ? channel : 0));
            	
            
            if (tossim) {
            	
                mote = motes.get(0);
            
            } else {
            
            	if (motes.containsKey(moteID)) {
                
            		mote = motes.get(moteID);
                
            	}
            }
            
            
            if (!sync) System.out.println("Sending command to: " + mote.name);
            
            
            mote.moteIF.send(MoteIF.TOS_BCAST_ADDR, cmd);

        } catch (IOException exception) {
        	
            System.err.println("Exception thrown when sending packets. Exiting.");
        
            System.err.println(exception);
            
            System.exit(1);
        
        }
    }

        
    private void closeLog() {
    	
        try {
        
        	out.close();
        
        } catch (IOException e) {
        
        	System.err.println("Problems closing log...");
        
        }
    }
    
    
    public synchronized void messageReceived (int to, Message message) {

    	if (message.amType() == ReportMessage.AM_TYPE) {
        	
        	ReportMessage msg = (ReportMessage) message;
        	
            
            try {
            	
                byte rssi = msg.get_rssi();

                // If this is not TOSSIM the RSSI values are direct from
                // hardware and therefore don't have the offset added to them
                // yet. Otherwise they do.
                if (!tossim) {
                	
                	rssi = (byte) (rssi + -45);
                	
                }

                
                out.write(
                	msg.get_recv()  + " " +
                	msg.get_src()   + " " +
                	msg.get_seqno() + " " +
                	rssi            + " " +
                	msg.get_lqi()   + "\n"
                );
                
                
                out.flush();
                
            } catch (IOException e) {
            	
                System.err.println("Problem printing the log");
            
            }

        } else {
        	
            PrintfMsg msg = (PrintfMsg) message;
            
            
            for (int i = 0, size = msg.dataLength(); i < size; i++) {
            
            	char nextChar = (char) msg.getElement_buffer(i);
                
            	System.out.print(nextChar);
            
            }
        }
    }


 
    private class Mote {
    	
    	public final String name;
    	public final MoteIF moteIF;

    	public Mote(String n, MoteIF m) {
    		name   = n;
    		moteIF = m;
    	}
    }

}
