import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.TreeMap;


public class Parser {

	private BufferedReader infile;

	private BufferedWriter out;
	private Node[] nodes;
	private boolean hasMap = false;
	private int[][] packets;
	private int[][] rssiSum;
	private int[][] lqiSum;
	private int senderColumn   = 0;
	private int receiverColumn = 1;
	private int rssiColumn     = 3;
	private int lqiColumn      = 4;
	private int numNodes = 0, iterations, payload, channel, power, width, height;
	private TreeMap<Integer, String> aliases = new TreeMap<Integer, String>();
	private TreeMap<String, Integer[]> coords = new TreeMap<String, Integer[]>();
	private ArrayList<Link> links = new ArrayList<Link>();
	
	
	
	public Parser(String profile, String outfile, String mapfile) {
		
		if (mapfile != null) {
			readMap(mapfile);
			hasMap = true;
		}
		
		openProfile(profile);
		processComments();
		
		packets = new int[numNodes][numNodes];
		rssiSum = new int[numNodes][numNodes];
		lqiSum  = new int[numNodes][numNodes];
		nodes   = new Node[numNodes];
		
		for (int i = 0; i < numNodes; i++) {
			nodes[i] = null;
		}
		
		processLines();
		closeInfile();
		openOutfile(outfile);
		processData();
		
	}
	
	
	public BufferedReader getInfile() {
		return infile;
	}


	public BufferedWriter getOut() {
		return out;
	}


	public Node[] getNodes() {
		return nodes;
	}


	public boolean isHasMap() {
		return hasMap;
	}


	public int[][] getPackets() {
		return packets;
	}


	public int[][] getRssiSum() {
		return rssiSum;
	}


	public int[][] getLqiSum() {
		return lqiSum;
	}


	public int getSenderColumn() {
		return senderColumn;
	}


	public int getReceiverColumn() {
		return receiverColumn;
	}


	public int getRssiColumn() {
		return rssiColumn;
	}


	public int getLqiColumn() {
		return lqiColumn;
	}


	public int getNumNodes() {
		return numNodes;
	}


	public int getIterations() {
		return iterations;
	}


	public int getPayload() {
		return payload;
	}


	public int getChannel() {
		return channel;
	}


	public int getPower() {
		return power;
	}


	public int getWidth() {
		return width;
	}


	public int getHeight() {
		return height;
	}


	public TreeMap<Integer, String> getAliases() {
		return aliases;
	}


	public TreeMap<String, Integer[]> getCoords() {
		return coords;
	}


	public ArrayList<Link> getLinks() {
		return links;
	}


	private void readMap(String map) {
		
		String line;
		
		
		try {
			
			infile = new BufferedReader(new FileReader(map));
			
			processComments();
			
			while ((line = infile.readLine()) != null) {

				if  (line.length() > 0 && !line.startsWith("#")) {
						
					String[] cols = line.split("\\s+");

					Integer[] pos = new Integer[2];
					pos[0] = Integer.parseInt(cols[1]);
					pos[1] = Integer.parseInt(cols[2]);
					
					coords.put(cols[0], pos);
						
			    }
			}
			
			closeInfile();
		
		} catch (FileNotFoundException e) {

			System.err.println("Unable to open map \"" + map + "\"");
			
			System.exit(1);
			
		} catch (IOException e) {

			System.err.println("Problem with mapfile...");
			
			System.exit(1);

		}
		
	}
	
	
	private void openProfile(String profile) {
		
		try {
			
			infile = new BufferedReader(new FileReader(profile));
		
		} catch (FileNotFoundException e) {

			System.err.println("Unable to open profile \"" + profile + "\"");
			
			System.exit(1);
		
		}
		
	}
	
	
	private void closeInfile() {
		
		try {
			
			infile.close();
		
		} catch (IOException e) {

			System.err.println("Problem closing infile...");
			
			System.exit(1);

		}
		
	}
	
	
	private void processComments() {
		
		String line;
		boolean isComment = true;
		
		
		try {
			
			while (isComment) {
				
				infile.mark(100);

				
				if ((line = infile.readLine()) == null) {
					
					System.err.println("End of file reached unexpectedly.");
					
					System.exit(1);
					
				}
				
				
				if  (line.length() > 0) {

					if (line.startsWith("#")) {
						
						parseComment(line);
						
					} else {
						
						infile.reset();
						
						break;
						
					}
			    }
			}
			
		} catch (IOException e) {

			System.err.println("Problem reading from infile! Unable to continue...");
			
			System.exit(1);

		}
		
	}
	
	
	private void parseComment(String comment) {
		
		String[] cols = comment.split("\\s+");
		
		
		try {

			 if (cols[0].equals("#alias")) {
		            
				 aliases.put(Integer.parseInt(cols[1]), cols[2]);
	                
	        } else if (cols[0].equals("#nodes")) {
	            
        		numNodes = Integer.parseInt(cols[1]);
                
        	} else if (cols[0].equals("#iterations")) {
            
        		iterations = Integer.parseInt(cols[1]);
                
        	} else if (cols[0].equals("#power")) {
            
        		power = Integer.parseInt(cols[1]);
                
        	} else if (cols[0].equals("#channel")) {
            
        		channel = Integer.parseInt(cols[1]);
                
        	} else if (cols[0].equals("#payload")) {
            
        		payload = Integer.parseInt(cols[1]);
                
        	} else if (cols[0].equals("#w")) {
            
        		width = Integer.parseInt(cols[1]);
                
        	} else if (cols[0].equals("#h")) {
            
        		height = Integer.parseInt(cols[1]);
                
        	}
        	
        } catch (ArrayIndexOutOfBoundsException e) {
        	
			System.err.println("Problem reading line: " + comment);
			
			System.exit(1);
        
        }
		
	}
	
	
	private void processLines() {
		
		String line;
		
		
		try {
			
			while ((line = infile.readLine()) != null) {

				if  (line.length() > 0 && !line.startsWith("#")) {
						
					parseEntry(line);
						
			    }
			}
			
		} catch (IOException e) {

			System.err.println("Problem reading from profile! Unable to continue...");
			
			System.exit(1);

		}
	}
	
	
	private void parseEntry(String entry) {

		String[] cols = entry.split("\\s+");
		
		
		try {
			
			int sender   = Integer.parseInt(cols[senderColumn]) - 1;
			int receiver = Integer.parseInt(cols[receiverColumn]) - 1;
			int rssi     = Integer.parseInt(cols[rssiColumn]);
			int lqi      = Integer.parseInt(cols[lqiColumn]);
			
			
			packets[sender][receiver] += 1;
			rssiSum[sender][receiver] += rssi;
			lqiSum [sender][receiver] += lqi;
			
			
		} catch (ArrayIndexOutOfBoundsException e) {
			
			System.err.println("Invalid column specification!");
			
			System.exit(1);
			
		}
	}

	
	private void openOutfile(String outfile) {
		
		try {
			   
            out = new BufferedWriter(new FileWriter(outfile));

            
    	} catch (IOException e) {
    		
            e.printStackTrace();
            
            System.exit(1);
            
        }
		
	}
	
	
	private void processData() {
		
		for (int n = 0; n < numNodes; n++) {
			String name = (aliases.containsKey(n + 1)) ?
				aliases.get(n + 1) :
				Integer.toString(n + 1);

			
			int x = 0;
			int y = 0;
			
			if (hasMap) {
				Integer[] pos = coords.get(name);
				x = pos[0];
				y = pos[1];
			}
			
			nodes[n] = new Node(name, n, x, y);
		}
		
		
		for (int i = 0; i < numNodes - 1; i++) {

			Node n1 = nodes[i];

			
			for (int j = i + 1; j < numNodes; j++) {
				
				
				Node n2 = nodes[j];

				
				Link l = new Link(
					n1, n2,
					iterations,
					packets[i][j],
					packets[j][i],
					rssiSum[i][j],
					rssiSum[j][i],
					lqiSum[i][j],
					lqiSum[j][i]
				);
				
				links.add(l);
				
				Node.setLink(n1, n2, l);
					
			}
		}
	}
	
}
